@@ -20,11 +20,11 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 
 import sys
 import os
-from util import logError, m18n, m18nc, isAlive, logWarning
-from common import WINDS, LIGHTSOURCES, InternalParameters, Preferences
+import datetime
+from util import logError, m18n, m18nc, logWarning, logDebug
+from common import WINDS, LIGHTSOURCES, Options, Internal, Preferences, isAlive
 import cgitb, tempfile, webbrowser
 from twisted.internet.defer import succeed, fail
-from twisted.python.failure import Failure
 
 class MyHook(cgitb.Hook):
     """override the standard cgitb hook: invoke the browser"""
@@ -45,37 +45,38 @@ try:
     from PyQt4.QtCore import Qt, QVariant, \
         QEvent, QMetaObject, PYQT_VERSION_STR, QString
     from PyQt4.QtGui import QPushButton, QMessageBox
-    from PyQt4.QtGui import QWidget, QColor, QBrush
+    from PyQt4.QtGui import QWidget
     from PyQt4.QtGui import QGridLayout, QAction
     from PyQt4.QtGui import QComboBox, QSlider, QHBoxLayout, QLabel
     from PyQt4.QtGui import QVBoxLayout, QSpacerItem, QSizePolicy, QCheckBox
-except ImportError, importError:
+except ImportError as importError:
     NOTFOUND.append('Package python-qt4: PyQt4: %s' % importError)
 
 try:
-    from zope.interface import implements # pylint: disable=W0611
-except ImportError, importError:
+    from zope.interface import implements # pylint: disable=unused-import
+except ImportError as importError:
     NOTFOUND.append('Package python-zope-interface missing: %s' % importError)
 
-from kde import Sorry, QuestionYesNo, KIcon, KAction, KApplication, KToggleFullScreenAction, \
+from kde import QuestionYesNo, KIcon, KAction, KApplication, KToggleFullScreenAction, \
     KXmlGuiWindow, KConfigDialog, KStandardAction
 
 try:
-    from query import Query
-    from tile import Tile
+    from query import Query, Transaction
     from board import WindLabel, FittingView, SelectorBoard, DiscardBoard, MJScene
-    from handboard import HandBoard
+    from handboard import ScoringHandBoard
     from playerlist import PlayerList
     from tileset import Tileset
     from background import Background
     from games import Games
     from statesaver import StateSaver
     from hand import Hand
-    from meld import Meld
+    from tile import Tile
+    from uitile import UITile
+    from meld import Meld, CONCEALED
     from scoring import ExplainView, ScoringDialog, ScoreTable
     from tables import SelectRuleset
     from client import Client
-    from humanclient import HumanClient, AlreadyConnected, LoginAborted, NetworkOffline
+    from humanclient import HumanClient
     from rulesetselector import RulesetSelector
     from tilesetselector import TilesetSelector
     from backgroundselector import BackgroundSelector
@@ -83,10 +84,10 @@ try:
     from uiwall import UIWall
     from animation import animate, afterCurrentAnimationDo, Animated
     from player import Player, Players
-    from game import Game, ScoringGame
+    from game import Game
     from chat import ChatWindow
 
-except ImportError, importError:
+except ImportError as importError:
     NOTFOUND.append('kajongg is not correctly installed: modules: %s' % importError)
 
 if len(NOTFOUND):
@@ -150,16 +151,13 @@ class ConfigDialog(KConfigDialog):
     """configuration dialog with several pages"""
     def __init__(self, parent, name):
         super(ConfigDialog, self).__init__(parent, QString(name), Preferences)
-        self.rulesetSelector = RulesetSelector(self)
-        self.pages = []
-        self.pages.append(self.addPage(PlayConfigTab(self),
-                m18nc('kajongg','Play'), "arrow-right"))
-        self.pages.append(self.addPage(TilesetSelector(self),
-                m18n("Tiles"), "games-config-tiles"))
-        self.pages.append(self.addPage(BackgroundSelector(self),
-                m18n("Backgrounds"), "games-config-background"))
-        self.pages.append(self.addPage(self.rulesetSelector,
-                m18n("Rulesets"), "games-kajongg-law"))
+        self.pages = [
+            self.addPage(PlayConfigTab(self),
+                m18nc('kajongg','Play'), "arrow-right"),
+            self.addPage(TilesetSelector(self),
+                m18n("Tiles"), "games-config-tiles"),
+            self.addPage(BackgroundSelector(self),
+                m18n("Backgrounds"), "games-config-background")]
         StateSaver(self)
 
     def keyPressEvent(self, event):
@@ -171,28 +169,6 @@ class ConfigDialog(KConfigDialog):
             return
         KConfigDialog.keyPressEvent(self, event)
 
-    def showEvent(self, dummyEvent):
-        """start transaction"""
-        self.rulesetSelector.refresh()
-        Query.dbhandle.transaction()
-
-    def accept(self):
-        """commit transaction"""
-        Query.dbhandle.commit() # commit renames and deletes of rulesets
-        if self.rulesetSelector.save():
-            KConfigDialog.accept(self)
-            return
-        Sorry(m18n('Cannot save your ruleset changes.<br>' \
-            'You probably introduced a duplicate name. <br><br >Message from database:<br><br>' \
-           '<message>%1</message>', Query.dbhandle.lastError().text()))
-
-    def reject(self):
-        """rollback transaction"""
-        self.rulesetSelector.cancel()
-        Query.dbhandle.rollback()
-        KConfigDialog.reject(self)
-
-
 class SwapDialog(QMessageBox):
     """ask the user if two players should change seats"""
     def __init__(self, swappers):
@@ -205,16 +181,10 @@ class SwapDialog(QMessageBox):
         self.noAnswer = QPushButton(m18n("&Keep seat"))
         self.addButton(self.noAnswer, QMessageBox.NoRole)
 
-    def exec_(self):
-        """I do not understand the logic of the exec return value. The yes button returns 0
-        and the no button returns 1. According to the C++ doc, the return value is an
-        opaque value that should not be used."""
-        return self.clickedButton() == self.yesAnswer
-
 class SelectPlayers(SelectRuleset):
-    """a dialog for selecting four players"""
+    """a dialog for selecting four players. Used only for scoring game."""
     def __init__(self, game):
-        SelectRuleset.__init__(self, game.client.host if game and game.client else None)
+        SelectRuleset.__init__(self)
         self.game = game
         Players.load()
         self.setWindowTitle(m18n('Select four players') + ' - Kajongg')
@@ -225,7 +195,6 @@ class SelectPlayers(SelectRuleset):
             cbName.manualSelect = False
             # increase width, we want to see the full window title
             cbName.setMinimumWidth(350) # is this good for all platforms?
-            # add all player names belonging to no host
             cbName.addItems(Players.humanNames.values())
             self.grid.addWidget(cbName, idx+1, 1)
             self.nameWidgets.append(cbName)
@@ -275,59 +244,74 @@ class SelectPlayers(SelectRuleset):
         self.names = list(unicode(cbName.currentText()) for cbName in self.nameWidgets)
         assert len(set(self.names)) == 4
 
-class VisiblePlayer(Player):
-    """this player instance has a visual representation"""
-    # pylint: disable=R0904
-    # too many public methods
-    def __init__(self, game, idx):
-        assert game
+class ScoringPlayer(Player):
+    """Player in a scoring game"""
+    # pylint: disable=too-many-public-methods
+    def __init__(self, game):
         self.handBoard = None # because Player.init calls clearHand()
-        Player.__init__(self, game)
-        self.idx = idx
-        self.front = game.wall[idx]
         self.manualRuleBoxes = []
-        self.handBoard = HandBoard(self)
-        self.handBoard.setVisible(False)
+        Player.__init__(self, game)
+        self.__front = self.game.wall[self.idx] # need front before setting handBoard
+        self.handBoard = ScoringHandBoard(self)
+
+    def hide(self):
+        """clear visible data and hide"""
+        self.clearHand()
+        self.handBoard.hide()
 
     def clearHand(self):
         """clears attributes related to current hand"""
-        self.manualRuleBoxes = []
         Player.clearHand(self)
+        if self.game and self.game.wall:
+            # is None while __del__
+            self.front = self.game.wall[self.idx]
+        if self.handBoard and isAlive(self.handBoard):
+            self.handBoard.setEnabled(True)
+            self.handBoard.showMoveHelper()
+        self.manualRuleBoxes = []
+
+    @property
+    def idx(self):
+        """our index in the player list"""
+        if not self in self.game.players:
+            # we will be added next
+            return len(self.game.players)
+        return self.game.players.index(self)
+
+    @property
+    def front(self):
+        """front"""
+        return self.__front
+
+    @front.setter
+    def front(self, value):
+        """also assign handBoard to front"""
+        self.__front = value
+        if value and self.handBoard:
+            self.handBoard.setParentItem(value)
+
+    @property
+    def handTotal(self):
+        """the hand total of this player"""
+        if self.hasManualScore():
+            spValue = Internal.field.scoringDialog.spValues[self.idx]
+            return spValue.value()
+        else:
+            return self.hand.total()
+
+    def handTotalForWall(self):
+        """returns the total for the current hand"""
+        return self.handTotal
 
     def hasManualScore(self):
         """True if no tiles are assigned to this player"""
-        if InternalParameters.field.scoringDialog:
-            return InternalParameters.field.scoringDialog.spValues[self.idx].isEnabled()
+        if Internal.field.scoringDialog:
+            return Internal.field.scoringDialog.spValues[self.idx].isEnabled()
         return False
 
-    def syncHandBoard(self, adding=None):
-        """update display of handBoard. Set Focus to tileName."""
-        self.handBoard.sync(adding)
-
-    def moveMeld(self, meld):
-        """a meld moves within our handBoard"""
-        assert meld.tiles[0].board == self.handBoard
-        self.removeMeld(meld)
-        self.addMeld(meld)
-
-    def colorizeName(self):
-        """set the color to be used for showing the player name on the wall"""
-        if self == self.game.activePlayer and self.game.client:
-            color = Qt.blue
-        elif InternalParameters.field.tilesetName == 'jade':
-            color = Qt.white
-        else:
-            color = Qt.black
-        self.front.nameLabel.setBrush(QBrush(QColor(color)))
-
-    def getsFocus(self, dummyResults=None):
-        """give this player focus on his handBoard"""
-        self.handBoard.setEnabled(True)
-        self.handBoard.hasFocus = True
-
     def refreshManualRules(self, sender=None):
         """update status of manual rules"""
-        assert InternalParameters.field
+        assert Internal.field
         if not self.handBoard:
             # might happen at program exit
             return
@@ -359,8 +343,7 @@ class VisiblePlayer(Player):
             wonChar = 'M'
         else:
             wonChar = 'm'
-        lastTile = InternalParameters.field.computeLastTile()
-        if lastTile and lastTile.istitle():
+        if self.lastTile and self.lastTile.istitle():
             lastSource = 'w'
         else:
             lastSource = 'd'
@@ -380,68 +363,168 @@ class VisiblePlayer(Player):
 
     def __lastString(self, asWinner):
         """compile hand info into a string as needed by the scoring engine"""
-        if not asWinner and self != self.game.winner:
+        if not asWinner or self != self.game.winner:
             return ''
-        lastTile = InternalParameters.field.computeLastTile()
-        if not lastTile:
+        if not self.lastTile:
             return ''
-        return 'L%s%s' % (lastTile, InternalParameters.field.computeLastMeld().joined)
+        return 'L%s%s' % (self.lastTile, self.lastMeld)
 
-    def computeHand(self, withTile=None, robbedTile=None, singleRule=None, asWinner=False):
+    def computeHand(self, singleRule=None, asWinner=None): # pylint: disable=arguments-differ
         """returns a Hand object, using a cache"""
-        game = self.game
-        if not game.isScoringGame():
-            # maybe we need a more extended class hierarchy for Player, VisiblePlayer, ScoringPlayer,
-            # PlayingPlayer but not now. Just make sure that ExplainView can always call the
-            # same computeHand regardless of the game type
-            return Player.computeHand(self, withTile=withTile, robbedTile=robbedTile, asWinner=asWinner)
-        if not self.handBoard:
-            return None
+        if asWinner is None:
+            asWinner = self == self.game.winner
+        self.lastTile = Internal.field.computeLastTile()
+        self.lastMeld = Internal.field.computeLastMeld()
         string = ' '.join([self.scoringString(), self.__mjstring(singleRule, asWinner), self.__lastString(asWinner)])
-        if game.eastMJCount == 8 and self == game.winner and self.wind == 'E':
-            cRules = [game.ruleset.findRule('XEAST9X')]
+        return Hand.cached(self, string, computedRules=singleRule)
+
+    def sortRulesByX(self, rules):
+        """if this game has a GUI, sort rules by GUI order of the melds they are applied to"""
+        withMelds = list(x for x in rules if x.meld)
+        withoutMelds = list(x for x in rules if x not in withMelds)
+        tuples = list(tuple([x, self.handBoard.findUIMeld(x.meld)]) for x in withMelds)
+        tuples = sorted(tuples, key=lambda x: x[1][0].sortKey())
+        return list(x[0] for x in tuples) + withoutMelds
+
+    def addMeld(self, meld):
+        """add meld to this hand in a scoring game"""
+        meld = Meld(meld)  # convert UITile to Tile
+        if len(meld) == 1 and meld[0].isBonus():
+            self._bonusTiles.append(meld[0])
+        elif meld.state == CONCEALED and not meld.isKong():
+            self._concealedMelds.append(meld)
         else:
-            cRules = []
-        if singleRule:
-            cRules.append(singleRule)
-        return Hand.cached(self, string, computedRules=cRules)
+            self._exposedMelds.append(meld)
+        self._hand = None
 
-    def popupMsg(self, msg):
-        """shows a yellow message from player"""
-        if msg != 'No Claim':
-            self.speak(msg.lower())
-            yellow = self.front.message
-            yellow.setText('  '.join([unicode(yellow.msg), m18nc('kajongg', msg)]))
-            yellow.setVisible(True)
+    def removeMeld(self, uiMeld):
+        """remove a meld from this hand in a scoring game"""
+        meld = Meld(uiMeld)
+        if len(meld) == 1 and meld[0].isBonus():
+            self._bonusTiles.remove(meld[0])
+        else:
+            popped = False
+            for melds in [self._concealedMelds, self._exposedMelds]:
+                for idx, myMeld in enumerate(melds):
+                    if myMeld == meld:
+                        melds.pop(idx)
+                        popped = True
+            if not popped:
+                logDebug('%s: %s.removeMeld did not find %s' % (self.name, self.__class__.__name__, meld), showStack=3)
+                logDebug('    concealed: %s' % self._concealedMelds)
+                logDebug('      exposed: %s' % self._exposedMelds)
+        self._hand = None
 
-    def hidePopup(self):
-        """hide the yellow message from player"""
-        if isAlive(self.front.message):
-            self.front.message.msg = ''
-            self.front.message.setVisible(False)
+    def syncHandBoard(self, adding=None):
+        """update display of handBoard. Set Focus to tileName."""
+        self.handBoard.sync()
 
-    def speak(self, text):
-        """speak if we have a voice"""
-        if self.voice:
-            self.voice.speak(text, self.front.rotation())
+class ScoringGame(Game):
+    """we play manually on a real table with real tiles and use
+    kajongg only for scoring"""
+    playerClass =  ScoringPlayer
+
+    def __init__(self, names, ruleset, gameid=None, client=None, wantedGame=None):
+        Game.__init__(self, names, ruleset, gameid=gameid, client=client, wantedGame=wantedGame)
+        field = Internal.field
+        field.selectorBoard.load(self)
+        self.prepareHand()
+        self.initHand()
+        for player in self.players:
+            player.clearHand()
+        Internal.field.adjustView()
+        Internal.field.updateGUI()
+        self.wall.decorate()
+
+    def close(self):
+        """log off from the server and return a Deferred"""
+        field = Internal.field
+        if isAlive(field):
+            field.setWindowTitle('Kajongg')
+        if field:
+            field.selectorBoard.tiles = []
+            field.selectorBoard.allSelectorTiles = []
+            if isAlive(field.centralScene):
+                field.centralScene.removeTiles()
+            for player in self.players:
+                player.hide()
+            if self.wall:
+                self.wall.hide()
+            field.game = None
+            field.updateGUI()
+            field.scoringDialog = None
+        return Game.close(self)
+
+    def prepareHand(self):
+        """prepare a scoring game hand"""
+        Game.prepareHand(self)
+        if self.finished():
+            self.close()
+        else:
+            selector = Internal.field.selectorBoard
+            selector.refill()
+            selector.hasFocus = True
+            self.wall.build()
+
+    @staticmethod
+    def isScoringGame():
+        """are we scoring a manual game?"""
+        return True
+
+    def saveStartTime(self):
+        """write a new entry in the game table with the selected players"""
+        Game.saveStartTime(self)
+        # for PlayingGame, this one is already done in Connection.__updateServerInfoInDatabase
+        known = Query('update server set lastruleset=? where url=?',
+            list([self.ruleset.rulesetId, Query.localServerName]))
+        if not known:
+            Query('insert into server(url,lastruleset) values(?,?)',
+                list([self.ruleset.rulesetId, Query.localServerName]))
+
+    def _setGameId(self):
+        """get a new id"""
+        if not self.gameid:
+            # a loaded game has gameid already set
+            self.gameid = self._newGameId()
+
+    def _mustExchangeSeats(self, pairs):
+        """filter: which player pairs should really swap places?"""
+        # pylint: disable=no-self-use
+        return list(x for x in pairs if Internal.field.askSwap(x))
+
+    def savePenalty(self, player, offense, amount):
+        """save computed values to database, update score table and balance in status line"""
+        scoretime = datetime.datetime.now().replace(microsecond=0).isoformat()
+        with Transaction():
+            Query("INSERT INTO SCORE "
+                "(game,penalty,hand,data,manualrules,player,scoretime,"
+                "won,prevailing,wind,points,payments, balance,rotated,notrotated) "
+                "VALUES(%d,1,%d,?,?,%d,'%s',%d,'%s','%s',%d,%d,%d,%d,%d)" % \
+                (self.gameid, self.handctr, player.nameid,
+                    scoretime, int(player == self.winner),
+                    WINDS[self.roundsFinished % 4], player.wind, 0,
+                    amount, player.balance, self.rotated, self.notRotated),
+                list([player.hand.string, offense.name]))
+        if Internal.field:
+            Internal.field.updateGUI()
 
 class PlayField(KXmlGuiWindow):
     """the main window"""
-    # pylint: disable=R0902
-    # pylint we need more than 10 instance attributes
+    # pylint: disable=too-many-instance-attributes
 
     def __init__(self):
         # see http://lists.kde.org/?l=kde-games-devel&m=120071267328984&w=2
-        InternalParameters.field = self
+        Internal.field = self
         self.game = None
         self.__startingGame = False
         self.ignoreResizing = 1
         super(PlayField, self).__init__()
         self.background = None
         self.showShadows = None
-        self.clientDialog = None
+        self._clientDialog = None
 
         self.playerWindow = None
+        self.rulesetWindow = None
         self.scoreTable = None
         self.explainView = None
         self.scoringDialog = None
@@ -454,6 +537,21 @@ class PlayField(KXmlGuiWindow):
         for action in self.toolBar().actions():
             if 'onfigure' in action.text():
                 action.setPriority(QAction.LowPriority)
+        if Options.host:
+            self.playGame()
+
+    @property
+    def clientDialog(self):
+        """wrapper: hide dialog when it is set to None"""
+        return self._clientDialog
+
+    @clientDialog.setter
+    def clientDialog(self, value):
+        """wrapper: hide dialog when it is set to None"""
+        if isAlive(self._clientDialog) and not value:
+            self._clientDialog.timer.stop()
+            self._clientDialog.hide()
+        self._clientDialog = value
 
     def sizeHint(self):
         """give the main window a sensible default size"""
@@ -495,7 +593,7 @@ class PlayField(KXmlGuiWindow):
         if self.scoringDialog:
             self.scoringDialog.slotInputChanged()
         if self.game and not self.game.finished():
-            self.game.wall.decoratePlayer(handBoard.player)
+            self.game.wall.decoratePlayer(handBoard.player) # pylint:disable=no-member
         # first decorate walls - that will compute player.handBoard for explainView
         if self.explainView:
             self.explainView.refresh(self.game)
@@ -529,6 +627,7 @@ class PlayField(KXmlGuiWindow):
         QGraphicsView.drawBackground always wants a pixmap
         for a huge rect like 4000x3000 where my screen only has
         1920x1200"""
+        # pylint: disable=too-many-statements
         self.setObjectName("MainWindow")
         centralWidget = QWidget()
         scene = MJScene()
@@ -558,8 +657,9 @@ class PlayField(KXmlGuiWindow):
         self.actionPlayGame = self.__kajonggAction("play", "arrow-right", self.playGame, Qt.Key_N)
         self.actionAbortGame = self.__kajonggAction("abort", "dialog-close", self.abortAction, Qt.Key_W)
         self.actionAbortGame.setEnabled(False)
-        self.actionQuit = self.__kajonggAction("quit", "application-exit", self.quit, Qt.Key_Q)
+        self.actionQuit = self.__kajonggAction("quit", "application-exit", self.close, Qt.Key_Q)
         self.actionPlayers = self.__kajonggAction("players", "im-user", self.slotPlayers)
+        self.actionRulesets = self.__kajonggAction("rulesets", "games-kajongg-law", self.slotRulesets)
         self.actionChat = self.__kajonggToggleAction("chat", "call-start",
             shortcut=Qt.Key_H, actionData=ChatWindow)
         game = self.game
@@ -583,108 +683,81 @@ class PlayField(KXmlGuiWindow):
         self.actionAutoPlay = self.__kajonggAction("demoMode", "arrow-right-double", None, Qt.Key_D)
         self.actionAutoPlay.setCheckable(True)
         self.actionAutoPlay.toggled.connect(self.__toggleDemoMode)
-        self.actionAutoPlay.setChecked(InternalParameters.autoPlay)
+        self.actionAutoPlay.setChecked(Internal.autoPlay)
         QMetaObject.connectSlotsByName(self)
 
     def showWall(self):
         """shows the wall according to the game rules (lenght may vary)"""
-        UIWall(self.game)
+        UIWall(self.game)   # sets self.game.wall
         if self.discardBoard:
             # scale it such that it uses the place within the wall optimally.
             # we need to redo this because the wall length can vary between games.
             self.discardBoard.maximize()
 
-    def genPlayers(self):
-        """generate four default VisiblePlayers"""
-        return Players([VisiblePlayer(self.game, idx) for idx in range(4)])
-
     def fullScreen(self, toggle):
         """toggle between full screen and normal view"""
         self.actionFullscreen.setFullScreen(self, toggle)
 
     def abortAction(self):
         """abort current game"""
-        try:
-            self.abort()
-        except Failure:
-            pass
+        def doNotQuit(dummy):
+            """ignore failure to abort"""
+        self.abort().addErrback(doNotQuit)
 
     def abort(self):
         """abort current game"""
-        def gotAnswer(result):
+        def gotAnswer(result, autoPlaying):
             """user answered"""
             if result:
                 return self.abortGame()
             else:
-                self.actionAutoPlay.setChecked(demoMode)
-                return fail(None) # just continue
+                self.actionAutoPlay.setChecked(autoPlaying)
+                return fail(Exception('no abort'))
+        def gotError(result):
+            """abortGame failed"""
+            logDebug('abortGame error:%s/%s ' % (str(result), result.getErrorMessage()))
         if not self.game:
             self.startingGame = False
             return succeed(None)
-        demoMode = self.actionAutoPlay.isChecked()
+        autoPlaying = self.actionAutoPlay.isChecked()
         self.actionAutoPlay.setChecked(False)
         if self.game.finished():
             return self.abortGame()
         else:
-            return QuestionYesNo(m18n("Do you really want to abort this game?"), gotAnswer)
-
-    def quit(self):
-        """exit the application"""
-        def doNotQuit(dummy):
-            """ignore failure to abort"""
-        deferred = self.abort()
-        deferred.addCallback(Client.shutdownClients).addCallback(Client.quitProgram).addErrback(doNotQuit)
-        return deferred
-
-    def hideGame(self, dummyResult=None):
-        """remove all visible traces of the current game"""
-        self.setWindowTitle('Kajongg')
-        self.discardBoard.hide()
-        self.selectorBoard.tiles = []
-        self.selectorBoard.allSelectorTiles = []
-        self.centralScene.removeTiles()
-        if self.clientDialog:
-            self.clientDialog.hide()
-            self.clientDialog = None
-        if self.game:
-            self.game.removeGameFromPlayfield()
-            self.game = None
-        self.updateGUI()
+            return QuestionYesNo(m18n("Do you really want to abort this game?"), always=True).addCallback(
+                gotAnswer, autoPlaying).addErrback(gotError)
 
     def abortGame(self):
         """if a game is active, abort it"""
-        self.actionAutoPlay.setChecked(False)
-        self.startingGame = False
         if self.game is None: # meanwhile somebody else might have aborted
             return succeed(None)
-        if self.game.isScoringGame():
-            self.hideGame()
-            return succeed(None)
-        else:
-            return self.game.close()
+        game = self.game
+        self.game = None
+        return game.close()
 
     def closeEvent(self, event):
         """somebody wants us to close, maybe ALT-F4 or so"""
-        if not self.quit():
-            event.ignore()
+        event.ignore()
+        def doNotQuit(dummy):
+            """ignore failure to abort"""
+        self.abort().addCallback(HumanClient.shutdownHumanClients).addCallbacks(Client.quitProgram, doNotQuit)
 
     def __moveTile(self, tile, wind, lowerHalf):
         """the user pressed a wind letter or X for center, wanting to move a tile there"""
         # this tells the receiving board that this is keyboard, not mouse navigation>
         # needed for useful placement of the popup menu
         assert self.game.isScoringGame()
-        assert isinstance(tile, Tile), (tile, str(tile))
+        assert isinstance(tile, UITile), (tile, str(tile))
         currentBoard = tile.board
-        dragTile, dragMeld = currentBoard.dragObject(tile)
         if wind == 'X':
             receiver = self.selectorBoard
         else:
             receiver = self.game.players[wind].handBoard
         if receiver != currentBoard or bool(lowerHalf) != bool(tile.yoffset):
-            movingLastMeld = tile.element in self.computeLastMeld().pairs
+            movingLastMeld = tile.tile in self.computeLastMeld()
             if movingLastMeld:
                 self.scoringDialog.clearLastTileCombo()
-            receiver.dropHere(dragTile, dragMeld, lowerHalf)
+            receiver.dropTile(tile, lowerHalf)
             if movingLastMeld and receiver == currentBoard:
                 self.scoringDialog.fillLastTileCombo()
 
@@ -695,17 +768,17 @@ class PlayField(KXmlGuiWindow):
         wind = chr(key%128)
         moveCommands = m18nc('kajongg:keyboard commands for moving tiles to the players ' \
             'with wind ESWN or to the central tile selector (X)', 'ESWNX')
-        tile = self.centralScene.focusItem().tile
+        uiTile = self.centralScene.focusItem()
         if wind in moveCommands:
             # translate i18n wind key to ESWN:
             wind = 'ESWNX'[moveCommands.index(wind)]
-            self.__moveTile(tile, wind, mod &Qt.ShiftModifier)
+            self.__moveTile(uiTile, wind, bool(mod &Qt.ShiftModifier))
             return True
         if key == Qt.Key_Tab and self.game:
             tabItems = [self.selectorBoard]
             tabItems.extend(list(p.handBoard for p in self.game.players if p.handBoard.tiles))
             tabItems.append(tabItems[0])
-            currentBoard = tile.board if isinstance(tile, Tile) else None
+            currentBoard = uiTile.board
             currIdx = 0
             while tabItems[currIdx] != currentBoard and currIdx < len(tabItems) -2:
                 currIdx += 1
@@ -743,6 +816,9 @@ class PlayField(KXmlGuiWindow):
         self.actionPlayers.setText(m18nc('@action:intoolbar', "&Players"))
         self.actionPlayers.setHelpText(m18nc('kajongg @info:tooltip', 'define your players.'))
 
+        self.actionRulesets.setText(m18nc('@action:intoolbar', "&Rulesets"))
+        self.actionRulesets.setHelpText(m18nc('kajongg @info:tooltip', 'customize rulesets.'))
+
         self.actionAngle.setText(m18nc('@action:inmenu', "&Change Visual Angle"))
         self.actionAngle.setIconText(m18nc('@action:intoolbar', "Angle"))
         self.actionAngle.setHelpText(m18nc('kajongg @info:tooltip', "Change the visual appearance of the tiles."))
@@ -782,6 +858,12 @@ class PlayField(KXmlGuiWindow):
             self.playerWindow = PlayerList(self)
         self.playerWindow.show()
 
+    def slotRulesets(self):
+        """show the player list"""
+        if not self.rulesetWindow:
+            self.rulesetWindow = RulesetSelector()
+        self.rulesetWindow.show()
+
     def selectScoringGame(self):
         """show all games, select an existing game or create a new game"""
         Players.load()
@@ -792,7 +874,7 @@ class PlayField(KXmlGuiWindow):
         if gameSelector.exec_():
             selected = gameSelector.selectedGame
             if selected is not None:
-                Game.loadFromDB(selected, what=ScoringGame)
+                ScoringGame.loadFromDB(selected)
             else:
                 self.newGame()
             if self.game:
@@ -809,19 +891,12 @@ class PlayField(KXmlGuiWindow):
     def playGame(self):
         """play a remote game: log into a server and show its tables"""
         self.startingGame = True
-        try:
-            HumanClient()
-        except AlreadyConnected:
-            pass
-        except LoginAborted:
-            pass
-        except NetworkOffline, exception:
-            Sorry(m18n('You have no network connection (error code %1).', str(exception)))
+        HumanClient()
 
     def adjustView(self):
         """adjust the view such that exactly the wanted things are displayed
         without having to scroll"""
-        if not InternalParameters.scaleScene:
+        if not Internal.scaleScene:
             return
         if self.game:
             with Animated(False):
@@ -840,42 +915,43 @@ class PlayField(KXmlGuiWindow):
         if oldRect != newRect:
             view.fitInView(scene.itemsBoundingRect(), Qt.KeepAspectRatio)
 
-    @apply
-    def startingGame(): # pylint: disable=E0202
+    @property
+    def startingGame(self):
         """are we trying to start a game?"""
-        # pylint: disable=W0212
-        def fget(self):
-            return self.__startingGame
-        def fset(self, value):
-            if value != self.__startingGame:
-                self.__startingGame = value
-                self.updateGUI()
-        return property(**locals())
+        return self.__startingGame
 
-    @apply
-    def tilesetName(): # pylint: disable=E0202
+    @startingGame.setter
+    def startingGame(self, value):
+        """are we trying to start a game?"""
+        if value != self.__startingGame:
+            self.__startingGame = value
+            self.updateGUI()
+
+    @property
+    def tilesetName(self):
         """the name of the current tileset"""
-        def fget(self):
-            return self.tileset.desktopFileName
-        def fset(self, name):
-            self.tileset = Tileset(name)
-        return property(**locals())
+        return self.tileset.desktopFileName
 
-    @apply
-    def backgroundName(): # pylint: disable=E0202
+    @tilesetName.setter
+    def tilesetName(self, name):
+        """the name of the current tileset"""
+        self.tileset = Tileset(name)
+
+    @property
+    def backgroundName(self):
         """setting this also actually changes the background"""
-        def fget(self):
-            return self.background.desktopFileName if self.background else ''
-        def fset(self, name):
-            """setter for backgroundName"""
-            self.background = Background(name)
-            self.background.setPalette(self.centralWidget())
-            self.centralWidget().setAutoFillBackground(True)
-        return property(**locals())
+        return self.background.desktopFileName if self.background else ''
+
+    @backgroundName.setter
+    def backgroundName(self, name):
+        """setter for backgroundName"""
+        self.background = Background(name)
+        self.background.setPalette(self.centralWidget())
+        self.centralWidget().setAutoFillBackground(True)
 
     def applySettings(self):
         """apply preferences"""
-        # pylint: disable=R0912
+        # pylint: disable=too-many-branches
         # too many branches
         self.actionAngle.setEnabled(bool(self.game) and Preferences.showShadows)
         animate() # drain the queue
@@ -969,7 +1045,7 @@ class PlayField(KXmlGuiWindow):
                 self.clientDialog.proposeAction() # an illegal action might have focus
                 self.clientDialog.selectButton() # select default, abort timeout
         else:
-            InternalParameters.autoPlay = checked
+            Internal.autoPlay = checked
             if checked:
                 # TODO: use the last used ruleset. Right now it always takes the first of the list.
                 self.playGame()
@@ -997,10 +1073,20 @@ class PlayField(KXmlGuiWindow):
         self.updateGUI()
         if self.game:
             self.game.wall.decorate()
+        if self.scoringDialog:
+            self.scoringDialog.clearLastTileCombo()
 
     def updateGUI(self):
         """update some actions, all auxiliary windows and the statusbar"""
+        if not isAlive(self):
+            return
+        title = ''
+        connections = list(x.connection for x in HumanClient.humanClients if x.connection)
         game = self.game
+        if not game:
+            title = ', '.join('{name}/{url}'.format(name=x.username, url=x.url) for x in connections)
+            if title:
+                self.setWindowTitle('%s - Kajongg' % title)
         for action in [self.actionScoreGame, self.actionPlayGame]:
             action.setEnabled(not bool(game))
         self.actionAbortGame.setEnabled(bool(game))
@@ -1017,7 +1103,7 @@ class PlayField(KXmlGuiWindow):
         self.actionChat.setChecked(self.actionChat.isEnabled() and bool(game.client.table.chatWindow))
         if self.actionScoring.isChecked():
             self.actionScoring.setChecked(scoring and not game.finished())
-        for view in [self.scoringDialog, self.explainView, self.scoreTable]:
+        for view in [self.explainView, self.scoreTable]:
             if view:
                 view.refresh(game)
         self.__showBalance()
@@ -1032,7 +1118,7 @@ class PlayField(KXmlGuiWindow):
         if self.game: # might be finished meanwhile
             with Animated(False):
                 wall = self.game.wall
-                oldIdx = LIGHTSOURCES.index(wall.lightSource)
+                oldIdx = LIGHTSOURCES.index(wall.lightSource) # pylint:disable=no-member
                 newLightSource = LIGHTSOURCES[(oldIdx + 1) % 4]
                 wall.lightSource = newLightSource
                 self.selectorBoard.lightSource = newLightSource
@@ -1077,4 +1163,7 @@ class PlayField(KXmlGuiWindow):
     def askSwap(swappers):
         """use this as a proxy such that module game does not have to import playfield.
         Game should also run on a server without KDE being installed"""
-        return SwapDialog(swappers).exec_()
+        return SwapDialog(swappers).exec_() == 0
+        # I do not understand the logic of the exec return value. The yes button returns 0
+        # and the no button returns 1. According to the C++ doc, the return value is an
+        # opaque value that should not be used."""
